from __future__ import annotations

import logging
from collections.abc import Callable, Iterable, Iterator
from itertools import tee
from typing import Any


def get_logger() -> logging.Logger:
    return logging.getLogger('kobuddy')


def unwrap[T](x: T | None) -> T:
    assert x is not None
    return x


# TODO switch to more_itertools?
def the[A](l: Iterable[A]) -> A:
    it = iter(l)
    try:
        first = next(it)
    except StopIteration as ee:
        raise RuntimeError('Empty iterator?') from ee
    assert all(e == first for e in it)
    return first


def group_by_key[T, K](l: Iterable[T], key: Callable[[T], K]) -> dict[K, list[T]]:
    res: dict[K, list[T]] = {}
    for i in l:
        kk = key(i)
        lst: list[T] = res.get(kk, [])
        lst.append(i)
        res[kk] = lst
    return res


type Res[V] = V | Exception


# TODO better name?
def split_res[V](it: Iterable[Res[V]]) -> tuple[Iterator[V], Iterator[Exception]]:
    vit, eit = tee(it)

    def it_val() -> Iterator[V]:
        for r in vit:
            if not isinstance(r, Exception):
                yield r

    def it_err() -> Iterator[Exception]:
        for r in eit:
            if isinstance(r, Exception):
                yield r

    return it_val(), it_err()


# TODO not sure if should keep it...
def sorted_res[V](it: Iterable[Res[V]], key: Callable[[V], Any]) -> Iterator[Res[V]]:
    vit, eit = split_res(it)
    yield from sorted(vit, key=key)
    yield from eit



# copy from HPI:my/core/sqlite.py
from __future__ import annotations

import sqlite3
from collections.abc import Callable, Iterator
from contextlib import contextmanager
from pathlib import Path
from typing import Any, Literal

SqliteRowFactory = Callable[[sqlite3.Cursor, sqlite3.Row], Any]


def dict_factory(cursor: sqlite3.Cursor, row: sqlite3.Row) -> dict[str, Any]:
    fields = [column[0] for column in cursor.description]
    return dict(zip(fields, row, strict=True))


Factory = SqliteRowFactory | Literal['row', 'dict']


@contextmanager
def sqlite_connection(
    db: Path | str, *, immutable: bool = False, row_factory: Factory | None = None
) -> Iterator[sqlite3.Connection]:
    dbp = f'file:{db}'
    # https://www.sqlite.org/draft/uri.html#uriimmutable
    if immutable:
        # assert results in nicer error than sqlite3.OperationalError
        assert Path(db).exists(), db
        dbp = f'{dbp}?immutable=1'
    row_factory_: Any = None
    if row_factory is not None:
        if callable(row_factory):
            row_factory_ = row_factory
        elif row_factory == 'row':
            row_factory_ = sqlite3.Row
        elif row_factory == 'dict':
            row_factory_ = dict_factory
        else:
            raise RuntimeError(f"Can't happen: {row_factory}")

    conn = sqlite3.connect(dbp, uri=True)
    try:
        conn.row_factory = row_factory_  # ty: ignore[invalid-assignment]
        with conn:
            yield conn
    finally:
        # Connection context manager isn't actually closing the connection, only keeps transaction
        conn.close()



from datetime import UTC, datetime
from pathlib import Path

import kobuddy


def get_test_db() -> Path:
    testdata = Path(__file__).absolute().parent.parent.parent.parent / 'testdata'
    assert testdata.exists(), testdata
    db = testdata / 'kobo_notes' / 'input' / 'KoboReader.sqlite'
    assert db.exists(), db
    return db


# a bit meh, but ok for now
kobuddy.set_databases(get_test_db())

from kobuddy import (
    _iter_events_aux,
    _iter_highlights,
    get_books_with_highlights,
    get_events,
)


def test_events() -> None:
    for e in _iter_events_aux():
        print(e)


def test_hls() -> None:
    for h in _iter_highlights():
        print(h)


def test_get_all() -> None:
    events = get_events()
    assert len(events) > 50
    for d in events:
        print(d)


def test_books_with_highlights() -> None:
    pages = get_books_with_highlights()

    g = pages[0]
    assert 'Essentialism' in g.book.title
    hls = g.highlights
    assert len(hls) == 273

    [b] = [h for h in hls if h.eid == '520b7b13-dbef-4402-9a81-0f4e0c4978de']
    # TODO wonder if there might be any useful info? StartContainerPath, EndContainerPath
    assert b.kind == 'bookmark'

    # TODO move to a more specific test?
    # TODO assert sorted by date or smth?
    assert hls[0].kind == 'highlight'
    # TODO assert highlights got no annotation? not sure if it's even necessary to distinguish..

    [ann] = [h for h in hls if len(h.annotation) > 0]

    assert ann.eid == 'eb264817-9a06-42fd-92ff-7bd38cd9ca79'
    assert ann.kind == 'annotation'
    assert (
        ann.text
        == 'He does this by finding which machine has the biggest queue of materials waiting behind it and finds a way to increase its efficiency.'
    )
    assert ann.annotation == 'Bottleneck'
    assert ann.dt == datetime(year=2017, month=8, day=12, hour=3, minute=49, second=13, microsecond=0, tzinfo=UTC)
    assert ann.book.author == 'Greg McKeown'

    assert len(pages) == 7


def test_history() -> None:
    kobuddy.print_progress()


def test_annotations() -> None:
    kobuddy.print_annotations()


def test_books() -> None:
    kobuddy.print_books()
